from .focal import FocalCosineLoss  # noqa: F401
from .online_smooth import OnlineLabelSmoothing  # noqa: F401
from .smooth import LabelSmoothingLoss  # noqa: F401
from .wrapper import CriterionMatcher, CriterionWrapper  # noqa: F401
